/*
 * Contributors :
 *   Pierre PETERLONGO, pierre.peterlongo@inria.fr [12/06/13]
 *   Nicolas MAILLET, nicolas.maillet@inria.fr     [12/06/13]
 *   Guillaume Collet, guillaume@gcollet.fr        [27/05/14]
 *
 * This software is a computer program whose purpose is to find all the
 * similar reads between two set of NGS reads. It also provide a similarity
 * score between the two samples.
 *
 * Copyright (C) 2014  INRIA
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "file_manager.h"
#include "file_manager_max.h"
#include "bloom_filter.h"
#include "boolean_vector.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <limits.h>
#include <iostream>
#include <string>
#include <ctime>
#include <map>

std::string version = "2.1";

// -----------------------------------------------------------------------
//                              PROTOTYPE
// -----------------------------------------------------------------------
void print_usage ();
unsigned long compare (FileManager * index_file_manager, FileManager * search_file_manager, int kmer_size, int min_hits, unsigned long max_kmer, std::ofstream & log_file);
unsigned long filter (FileManager * file_manager, int min_size, int max_N, float min_shannon);

// -----------------------------------------------------------------------
//                                MAIN
// -----------------------------------------------------------------------

int main (int argc, char ** argv)
{
	const clock_t begin_time = clock();
	
	FileManager * A_file_manager;
	FileManager * B_file_manager;
	
	std::map<std::string, std::string> A_files;
	std::map<std::string, std::string> B_files;
	
	int kmer_size = 32;
	int min_hits = 2;
	unsigned long max_nb_reads = 0;
	unsigned long max_kmer = (unsigned long) (1000000000.0 / pow (2, 33 - kmer_size));
	
	std::string log_path;
	std::string out_path;
	
	int min_size = kmer_size * min_hits;
	int max_N = INT_MAX;
	float min_shannon = 0.0;
	
	////////////////////////////////////////////////////////////
	// Read command line arguments
	//
	std::string tmp;
	int arg_pos = 1;
	while (arg_pos < argc){
		std::string flag = argv[arg_pos];
		if (flag.compare("-a") == 0) {
			// File to index name, optionaly followed by a boolean vector file after a comma
			arg_pos++;
			if (arg_pos >= argc) {
				std::cerr << "Error, flag " << argv[arg_pos - 1] << " needs an argument\n";
				print_usage();
				exit(1);
			}
			tmp = argv[arg_pos];
			if (tmp.find(',') < tmp.size()) {
				A_files[tmp.substr(0, tmp.find(','))] = tmp.substr(tmp.find(',') + 1);
				std::cout << "file A: " << tmp.substr(0, tmp.find(',')) << "\n";
			} else {
				A_files[tmp] = "";
				std::cout << "file A: " << tmp << "\n";
			}
		} else if (flag.compare("-b") == 0) {
			// File to search name, optionaly followed by a boolean vector file after a comma
			arg_pos++;
			if (arg_pos >= argc) {
				std::cerr << "Error, flag " << argv[arg_pos - 1] << " needs an argument\n";
				print_usage();
				exit(1);
			}
			tmp = argv[arg_pos];
			if (tmp.find(',') < tmp.size()) {
				B_files[tmp.substr(0, tmp.find(','))] = tmp.substr(tmp.find(',') + 1);
				std::cout << "file B: " << tmp.substr(0, tmp.find(',')) << "\n";
			} else {
				B_files[tmp] = "";
				std::cout << "file B: " << tmp << "\n";
			}
		} else if (flag.compare("-l") == 0) {
			// The path to the log directory
			arg_pos++;
			if (arg_pos >= argc) {
				std::cerr << "Error, flag " << argv[arg_pos - 1] << " needs an argument\n";
				print_usage();
				exit(1);
			}
			log_path = argv[arg_pos];
		} else if (flag.compare("-o") == 0) {
			// The path to the ouput directory
			arg_pos++;
			if (arg_pos >= argc) {
				std::cerr << "Error, flag " << argv[arg_pos - 1] << " needs an argument\n";
				print_usage();
				exit(1);
			}
			out_path = argv[arg_pos];
		} else if (flag.compare("-k") == 0) {
			// The size of k-mers
			arg_pos++;
			if (arg_pos >= argc) {
				std::cerr << "Error, flag " << argv[arg_pos - 1] << " needs an argument\n";
				print_usage();
				exit(1);
			}
			kmer_size = atoi(argv[arg_pos]);
			max_kmer = (unsigned long) (1000000000.0 / pow (2, 33 - kmer_size));
			std::cout << "k-mer size (-k) = " << kmer_size << "\n";
		} else if (flag.compare("-t") == 0) {
			// The minimal number of hits to consider two reads as similar
			arg_pos++;
			if (arg_pos >= argc) {
				std::cerr << "Error, flag " << argv[arg_pos - 1] << " needs an argument\n";
				print_usage();
				exit(1);
			}
			min_hits = atoi(argv[arg_pos]);
			std::cout << "min hits (-t) = " << min_hits << "\n";
		} else if (flag.compare("-m") == 0) {
			// The maximum number of reads per file (m first reads)
			arg_pos++;
			if (arg_pos >= argc) {
				std::cerr << "Error, flag " << argv[arg_pos - 1] << " needs an argument\n";
				print_usage();
				exit(1);
			}
			max_nb_reads = atoi(argv[arg_pos]);
			std::cout << "max reads (-m) = " << max_nb_reads << "\n";
		} else if (flag.compare("-r") == 0) {
			// The minimum size of reads
			arg_pos++;
			if (arg_pos >= argc) {
				std::cerr << "Error, flag " << argv[arg_pos - 1] << " needs an argument\n";
				print_usage();
				exit(1);
			}
			min_size = atoi (argv[arg_pos]);
		} else if (flag.compare("-n") == 0) {
			// The maximal number of Ns a read
			arg_pos++;
			if (arg_pos >= argc) {
				std::cerr << "Error, flag " << argv[arg_pos - 1] << " needs an argument\n";
				print_usage();
				exit(1);
			}
			max_N = atoi (argv[arg_pos]);
		} else if (flag.compare("-e") == 0) {
			// The minimal Shannon index
			arg_pos++;
			if (arg_pos >= argc) {
				std::cerr << "Error, flag " << argv[arg_pos - 1] << " needs an argument\n";
				print_usage();
				exit(1);
			}
			min_shannon = atof(argv[arg_pos]);
		} else if (flag.compare("-h") == 0) {
			print_usage ();
			return 0;
		} else if (flag.compare("-v") == 0) {
			std::cout << "\ncompare_reads version " << version << "\n";
			return 0;
		} else {
			std::cerr << "Unknown option " << flag << "\n";
            print_usage ();
			return (0);
		}
		arg_pos++;
	}
	
	////////////////////////////////////////////////////////////
	// Check the existence of at least one file to index and
	// one file to search
	//
	if (A_files.empty())
	{
		fprintf(stderr,"No file specified with -a\n");
		print_usage ();
		return 1;
	}
	
	if (B_files.empty())
	{
		fprintf(stderr,"No file specified with -b\n");
		print_usage ();
		return 1;
	}
	
	if (max_nb_reads > 0) {
		A_file_manager = new FileManagerMax (max_nb_reads);
		B_file_manager = new FileManagerMax (max_nb_reads);
	} else {
		A_file_manager = new FileManager ();
		B_file_manager = new FileManager ();
	}
	
	for (std::map<std::string, std::string>::iterator it = A_files.begin(); it != A_files.end(); it++) {
		if (it->second.empty()) {
			A_file_manager->addFile(it->first);
		} else {
			A_file_manager->addFile(it->first, it->second);
		}
	}
	for (std::map<std::string, std::string>::iterator it = B_files.begin(); it != B_files.end(); it++) {
		if (it->second.empty()) {
			B_file_manager->addFile(it->first);
		} else {
			B_file_manager->addFile(it->first, it->second);
		}
	}
	
	////////////////////////////////////////////////////////////
	// If out_path is empty then the directory
	// of the first index file is used
	//
	if (out_path.empty()) {
		if (A_files.begin()->first.rfind("/") < A_files.begin()->first.size()) {
			out_path = A_files.begin()->first.substr(0, A_files.begin()->first.rfind("/"));
		} else {
			out_path = ".";
		}
	}
	
	////////////////////////////////////////////////////////////
	// Check existence of log_path and out_path
	// create them if needed
	//
	if (log_path.empty()) {
		log_path = ".";
	}
	struct stat info;
	if (stat (log_path.c_str(), &info ) != 0){
		mkdir(log_path.c_str(), S_IRWXU|S_IRGRP|S_IXGRP);
	} else if (!(info.st_mode & S_IFDIR)){
		std::cerr << "Error: " << log_path << " already exists and is not a directory\n";
		exit(1);
	}
	
	if (stat (out_path.c_str(), &info ) != 0){
		mkdir(out_path.c_str(), S_IRWXU|S_IRGRP|S_IXGRP);
	} else if (!(info.st_mode & S_IFDIR)) {
		std::cerr << "Error: " << out_path << " already exists and is not a directory\n";
		exit(1);
	}
	
	const std::vector<std::string> & A_file_names = A_file_manager->get_file_names();
	const std::vector<std::string> & B_file_names = B_file_manager->get_file_names();
	
	////////////////////////////////////////////////////////////
	// Filter reads
	//
	std::cout << "\n------------------------------------------------------------------\n";
	std::cout << "filter reads\n";
	std::cout << "------------------------------------------------------------------\n";
	
	const clock_t begin_filter = clock();
	std::cout << "Filter reads from A : ";
	unsigned long nb_valid_A = filter (A_file_manager, min_size, max_N, min_shannon);
	std::cout << nb_valid_A << " valid reads\nFilter reads from B : ";
	unsigned long nb_valid_B = filter (B_file_manager, min_size, max_N, min_shannon);
	std::cout << nb_valid_B << " valid reads\n";
	std::cout << "Filter time: " << float(clock() - begin_filter) / CLOCKS_PER_SEC << " s\n";
	
	unsigned long nb_total_reads = A_file_manager->get_total_nb_reads() + B_file_manager->get_total_nb_reads();
	
	unsigned long nb_common = 0;

	std::ofstream log_file;
	std::string log_file_name;
	
	////////////////////////////////////////////////////////////
	// First step : B in A
	//
	
	// output message
	std::cout << "------------------------------------------------------------------\n";
	std::cout << "finding reads from {";
	int i = 0;
	for (; i < (int) B_file_names.size() - 1; i++) {
		std::cout << B_file_names.at(i).substr(B_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cout << B_file_names.at(i).substr(B_file_names.at(i).rfind('/') + 1) << "} present in raw {";
	i = 0;
	for (; i < (int) A_file_names.size() - 1; i++) {
		std::cout << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cout << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << "}\n";
	std::cout << "------------------------------------------------------------------\n";
	
	////////////////////////////////////////////////////////////
	// Generate the name of log file from log_path (-l)
	// Open log_file and write log header
	//
	log_file_name = log_path + "/" + A_file_manager->get_sum_of_file_names() + "_VS_" + B_file_manager->get_sum_of_file_names() + ".txt";
	log_file.open(log_file_name.c_str());
	if (!log_file.good()) {
		std::cerr << "Cannot open file " << log_file_name << " for writing\n";
	}
	log_file << "Compareads " << version << "\n";
	log_file << "Two reads (of at least " << min_size << " nucleotides and a Shannon index greater than " << min_shannon << ") are similar if they share at least " << min_hits << " k-mers of " << kmer_size << " nucleotides\n";
	log_file.precision(5);
	log_file << "Parameters: t=" << min_hits << " -- k=" << kmer_size << " (using " << (float) pow(2, kmer_size - 1)/(float) 1073741824 << " Gb), Max kmers in BDS=" << max_kmer << " \n";
	log_file << "Total number of reads in " << B_file_manager->get_sum_of_file_names() << ": " << B_file_manager->get_total_nb_reads() << "\n";
	if (max_nb_reads > 0 && max_nb_reads < B_file_manager->get_total_nb_reads()) {
		log_file << "* " << max_nb_reads << "reads maximum per files are used (option -m).\n";
	} else {
		log_file << "* " << B_file_manager->get_total_nb_reads() << " reads per files are used (all reads).\n";
	}
	
	// Compare first step
	compare (A_file_manager, B_file_manager, kmer_size, min_hits, max_kmer, log_file);
	log_file.close();
	
	////////////////////////////////////////////////////////////
	// Second step A in [B in A]
	//
	B_file_manager->apply_bv_on_files();
	A_file_manager->rewind();
	B_file_manager->rewind();
	
	// output message
	std::cout << "------------------------------------------------------------------\n";
	std::cout << "finding reads from {";
	i = 0;
	for (; i < (int) A_file_names.size() - 1; i++) {
		std::cout << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cout << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << "} present in [reads from {";
	i = 0;
	for (; i < (int) B_file_names.size() - 1; i++) {
		std::cout << B_file_names.at(i).substr(B_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cout << B_file_names.at(i).substr(B_file_names.at(i).rfind('/') + 1) << "} present in raw {";
	i = 0;
	for (; i < (int) A_file_names.size() - 1; i++) {
		std::cout << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cout << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << "}]\n";
	std::cout << "------------------------------------------------------------------\n";
	
	////////////////////////////////////////////////////////////
	// Generate the name of log file from log_path (-l)
	// Open log_file and write log header
	//
	log_file_name = log_path + "/" + B_file_manager->get_sum_of_file_names() + "_VS_" + A_file_manager->get_sum_of_file_names() + ".txt";
	log_file.open(log_file_name.c_str());
	if (!log_file.good()) {
		std::cerr << "Cannot open file " << log_file_name << " for writing\n";
	}
	log_file << "Compareads " << version << "\n";
	log_file << "Two reads (of at least " << min_size << " nucleotides and a Shannon index greater than " << min_shannon << ") are similar if they share at least " << min_hits << " k-mers of " << kmer_size << " nucleotides\n";
	log_file.precision(5);
	log_file << "Parameters: t=" << min_hits << " -- k=" << kmer_size << " (using " << (float) pow(2, kmer_size - 1)/(float) 1073741824 << " Gb), Max kmers in BDS=" << max_kmer << " \n";
	log_file << "Total number of reads in " << A_file_manager->get_sum_of_file_names() << ": " << A_file_manager->get_total_nb_reads() << "\n";
	if (max_nb_reads > 0 && max_nb_reads < A_file_manager->get_total_nb_reads()) {
		log_file << "* " << max_nb_reads << "reads maximum per files are used (option -m).\n";
	} else {
		log_file << "* " << A_file_manager->get_total_nb_reads() << " reads per files are used (all reads).\n";
	}

	// Compare second step
	nb_common += compare (B_file_manager, A_file_manager, kmer_size, min_hits, max_kmer, log_file);
	A_file_manager->save_bv(out_path, B_file_manager->get_sum_of_file_names());
	log_file.close();
	
	////////////////////////////////////////////////////////////
	// third step [B in A] in [A in [B in A]]
	//
	A_file_manager->apply_bv_on_files();
	A_file_manager->rewind();
	B_file_manager->rewind();
	
	// ouput message
	std::cout << "------------------------------------------------------------------\n";
	std::cout << "finding [reads from {";
	i = 0;
	for (; i < (int) B_file_names.size() - 1; i++) {
		std::cout << B_file_names.at(i).substr(B_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cout << B_file_names.at(i).substr(B_file_names.at(i).rfind('/') + 1) << "} present in raw {";
	i = 0;
	for (; i < (int) A_file_names.size() - 1; i++) {
		std::cout << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cout << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << "}] present in [reads from {";
	i = 0;
	for (; i < (int) A_file_names.size() - 1; i++) {
		std::cout << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cout << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << "} present in [reads from {";
	i = 0;
	for (; i < (int) B_file_names.size() - 1; i++) {
		std::cout << B_file_names.at(i).substr(B_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cout << B_file_names.at(i).substr(B_file_names.at(i).rfind('/') + 1) << "} present in raw {";
	i = 0;
	for (; i < (int) A_file_names.size() - 1; i++) {
		std::cout << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cout << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << "}]]\n";
	std::cout << "------------------------------------------------------------------\n";
	
	// Compare third step
	nb_common += compare (A_file_manager, B_file_manager, kmer_size, min_hits, max_kmer, log_file);
	B_file_manager->save_bv(out_path, A_file_manager->get_sum_of_file_names());
	
	std::cout << "\n-------------------------------------------\n";
	std::cout << "COMPUTATION DONE FOR {";
	i = 0;
	for (; i < (int) A_file_names.size() - 1; i++) {
		std::cout << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cout << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << "} VS {";
	i = 0;
	for (; i < (int) B_file_names.size() - 1; i++) {
		std::cout << B_file_names.at(i).substr(B_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cout << B_file_names.at(i).substr(B_file_names.at(i).rfind('/') + 1) << "}\n";
	
	std::cout << "Writing result files (may be long)\n";
	A_file_manager->save_files(out_path, B_file_manager->get_sum_of_file_names());
	B_file_manager->save_files(out_path, A_file_manager->get_sum_of_file_names());
	
	/* Write end message */
	std::cout << "Results are in\n";
	for (std::map<std::string, std::string>::iterator it = A_files.begin(); it != A_files.end(); it++) {
		std::cout << "\t" << out_path << "/" << it->first.substr(it->first.rfind('/') + 1) << "_in_" << B_file_manager->get_sum_of_file_names() << "\n";
	}
	std::cout << "and\n";
	for (std::map<std::string, std::string>::iterator it = B_files.begin(); it != B_files.end(); it++) {
		std::cout << "\t" << out_path << "/" << it->first.substr(it->first.rfind('/') + 1) << "_in_" << A_file_manager->get_sum_of_file_names() << "\n";
	}
	
	std::cout << "\nLog files are in\n";
	std::cout << "\t" << log_path << "/" << A_file_manager->get_sum_of_file_names() << "_VS_" << B_file_manager->get_sum_of_file_names() << ".txt\n";
	std::cout << "and\n";
	std::cout << "\t" << log_path << "/" << B_file_manager->get_sum_of_file_names() << "_VS_" << A_file_manager->get_sum_of_file_names() << ".txt\n";
	std::cout << "-------------------------------------------\n";
	std::cout << "Total time: " << float(clock () - begin_time) / CLOCKS_PER_SEC << " s\n";
	
	std::cout << "Similarity score between {";
	i = 0;
	for (; i < (int) B_file_names.size() - 1; i++) {
		std::cout << B_file_names.at(i).substr(B_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cout << B_file_names.at(i).substr(B_file_names.at(i).rfind('/') + 1) << "} and {";
	i = 0;
	for (; i < (int) A_file_names.size() - 1; i++) {
		std::cout << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cout << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << "}: " << (double) nb_common * 100.0 / (double) nb_total_reads << "%\n\n";
	
	return 0;
}

unsigned long compare (FileManager * index_file_manager, FileManager * search_file_manager, int kmer_size, int min_hits, unsigned long max_kmer, std::ofstream & log_file)
{
	// -----------------------------------------------------------------------
	//                           INDEX & SEARCH
	// -----------------------------------------------------------------------
	BloomFilter bloom_filter (kmer_size);
	HashKey hash (kmer_size);
	std::string & current_read_to_index = index_file_manager->get_next_read_to_compare();
	clock_t index_time = 0;
	clock_t search_time = 0;
	unsigned long nb_indexed_reads = 0;
	unsigned long nb_searched_reads = 0;
	unsigned long nb_found_reads = 0;
	while (!current_read_to_index.empty()) {
		// -----------------------------------------------------------------------
		//                                INDEX
		// -----------------------------------------------------------------------
		// Index reads from index_file_manager
		bloom_filter.clear();
		const clock_t begin_index = clock();
		unsigned long nb_indexed_kmers = 0;
		while (!current_read_to_index.empty() && nb_indexed_kmers < max_kmer) {
			nb_indexed_reads++;
			hash.clear();
			for (int i = 0; i < (int) current_read_to_index.size(); i++) {
				if (current_read_to_index[i] == 'N' || current_read_to_index[i] == 'n') {
					hash.clear();
				} else if (hash.add(current_read_to_index[i]) >= kmer_size) {
					bloom_filter.feed(hash);
					nb_indexed_kmers++;
				}
			}
			current_read_to_index = index_file_manager->get_next_read_to_compare();
		}
		const clock_t end_index = clock();
		index_time += end_index - begin_index;
		
		// -----------------------------------------------------------------------
		//                               SEARCH
		// -----------------------------------------------------------------------
		// Search reads from search_file_manager in the indexed reads
		nb_searched_reads = 0;
		search_file_manager->rewind();
		const clock_t begin_search = clock();
		std::string & current_read_to_search = search_file_manager->get_next_read_to_compare();
		while (!current_read_to_search.empty()) {
			nb_searched_reads++;
			while (!current_read_to_search.empty() && search_file_manager->is_tagged_current_read()) {
				current_read_to_search = search_file_manager->get_next_read_to_compare();
				nb_searched_reads++;
			}
			// Search k-mers
			int seen = 0;
			//search_file_manager->untag_current_read();
			bool found = false;
			hash.clear();
			for (long i = 0; i < (int) current_read_to_search.size() && !found; i++) {
				if (current_read_to_search[i] == 'N' || current_read_to_search[i] == 'n') {
					hash.clear();
				} else if (hash.add(current_read_to_search[i]) >= kmer_size) {
					if (bloom_filter.is_found(hash)) {
						seen++;
						if (seen >= min_hits) {
							search_file_manager->tag_current_read();
							nb_found_reads++;
							found = true;
						}
						hash.clear();
					}
				}
			}
			// Search the reverse strand if not found in the first step
			if (!found) {
				seen = 0;
				hash.clear();
				for (long i = 0; i < (int) current_read_to_search.size() && !found; i++) {
					if (current_read_to_search[i] == 'N' || current_read_to_search[i] == 'n') {
						hash.clear();
					} else if (hash.rv_add(current_read_to_search[i]) >= kmer_size) {
						if (bloom_filter.is_found(hash)) {
							seen++;
							if (seen >= min_hits) {
								search_file_manager->tag_current_read();
								nb_found_reads++;
								found = true;
							}
							hash.clear();
						}
					}
				}
			}
			current_read_to_search = search_file_manager->get_next_read_to_compare();
		}
		const clock_t end_search = clock();
		search_time += end_search - begin_search;
	}
	
	log_file << "[" << nb_indexed_reads << " reads indexed, " << nb_searched_reads << " reads searched, " << nb_found_reads << " reads shared]\n";
	log_file << "--- " << nb_found_reads << " (" << (float) nb_found_reads * 100 / (float)nb_searched_reads << "%)\n";
	std::cerr << "[" << nb_indexed_reads << " reads indexed, " << nb_searched_reads << " reads searched, " << nb_found_reads << " reads shared]\n";
	
	std::cerr << "Index time: " << float(index_time) / CLOCKS_PER_SEC << " s\n";
	std::cerr << "Query time: " << float(search_time) / CLOCKS_PER_SEC << " s\n";
	std::cerr << "Total time: " << float(index_time + search_time) / CLOCKS_PER_SEC << " s\n";

	return nb_found_reads;
}



// -----------------------------------------------------------------------
//                             PRINT USAGE
// -----------------------------------------------------------------------
void print_usage ()
{
	std::cerr << "\nCompare_reads, version " << version << "\n";
	std::cerr << "Usage : ./compare_reads -a <readsA.fasta>[,readsA.bv] -b <readsB.fastq>[,readsB.bv] [options]\n";
	std::cerr << "Mandatory:\n";
	std::cerr << "\t -a <readsA.fasta>[,readsA.bv]: The sequences to index (MUST be in fasta or fastq format)      - MANDATORY\n";
	std::cerr << "\t                                readsA.bv is the boolean vector of valid reads in readsA.fasta - OPTIONAL\n";
	std::cerr << "\t -b <readsB.fastq>[,readsB.bv]: The sequences to search (MUST be in fasta or fastq format)     - MANDATORY\n";
	std::cerr << "\t                                readsB.bv is the boolean vector of valid reads in readsB.fasta - OPTIONAL\n";
	std::cerr << "Options:\n";
	std::cerr << "\t -l </.../>: ABSOLUTE path to log folder\n";
	std::cerr << "\t -o </.../>: ABSOLUTE path to output folder\n";
	std::cerr << "\t -k <value>: Size of k-mers (value of k). [default=32]\n";
	std::cerr << "\t -t <value>: Number of shared k-mers. [default=2]\n";
	std::cerr << "\t -m <value>: Maximum number of reads to use in each files - [default=all]\n";
	std::cerr << "\t -r int\t\t: minimal length a read should have to be kept. [default=64]\n";
    std::cerr << "\t -n int\t\t: maximal number of Ns a read should contain to be kept. [default=any]\n";
    std::cerr << "\t -e float\t: minimal Shannon index a read should have to be kept. [default=0]\n";
    std::cerr << "\t -h: Prints this message and exit\n";
	std::cerr << "\t -v: Prints the version number and exit\n";
}

// -----------------------------------------------------------------------
//                        NUMBER OF Ns IN A READ
// -----------------------------------------------------------------------

int number_of_N (const std::string & read)
{
	int nb = 0;
	for (int i = 0; i < (int) read.size(); i++) {
		if (read[i] == 'N' || read[i] == 'n') {
			nb++;
		}
	}
	return nb;
}

// -----------------------------------------------------------------------
//                        SHANNON INDEX OF A READ
// -----------------------------------------------------------------------

float shannon_index(std::string & read)
{
	float index = 0;
	float freq [5];
	
	// Initiate frequencies to 0.0
	for (int i = 0; i < 5; i++) {
		freq[i] = 0.0;
	}
	
	// Frequency of each letter (A, C, G, T or N)
	for (int i = 0; i < (int) read.size(); i++)
	{
		switch (toupper(read[i]))
		{
			case 'A':
				freq[0] += 1.0;
				break;
			case 'C':
				freq[1] += 1.0;
				break;
			case 'G':
				freq[2] += 1.0;
				break;
			case 'T':
				freq[3] += 1.0;
				break;
			case 'N':
				freq[4] += 1.0;
				break;
			default:
				std::cerr << "Error in Shannon index. Unknown char '" << read[i] << "'. Must be A, C, G, T or N.\n";
				exit(1);
		}
	}
	
	// Shannon index calculation
	for (int i = 0; i < 5; i++)
	{
		freq[i] = freq[i] / (float) read.size();
		if (freq[i] != 0) {
			index += freq[i] * log (freq[i]) / log(2);
		}
	}
	return fabs (index);
}

unsigned long filter (FileManager * file_manager, int min_size, int max_N, float min_shannon)
{
	unsigned long nb_valid = 0;
	std::string & current_read = file_manager->get_next_read_to_filter();
	while (!current_read.empty()) {
		if ((int) current_read.size() >= min_size && number_of_N (current_read) <= max_N && shannon_index(current_read) >= min_shannon) {
			nb_valid++;
			file_manager->tag_current_read();
		}
		current_read = file_manager->get_next_read_to_filter();
	}
	file_manager->apply_bv_on_files();
	file_manager->rewind();
	return nb_valid;
}
